/*============================================================================
  File:     TableValued.sql

  Summary:  This script demonstrates the savings in database roundtrips
			gained by using table-valued parameters vs multiple calls.

  Date:     August 2008

  SQL Server Version: 10.0.1600.22 (RTM)
------------------------------------------------------------------------------
  Written by Bob Beauchemin, SQLskills.com

  For more scripts and sample code, check out http://www.SQLskills.com

  This script is intended only as a supplement to SQL Server Jumpstart program.
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

if DATABASEPROPERTY('testdb', 'Version') is null
create database testdb
go

use testdb
go

-- 
-- Step 1:
-- Create tables for orders and order items
--

IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPOrd'), 'IsTable') = 1
  drop table dbo.TVPOrd
GO

create table dbo.TVPOrd (
  CustCode varchar(5),
  OrdNo int identity,
  OrdDate datetime
)
go

IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPItem'), 'IsTable') = 1
  drop table dbo.TVPItem
GO

create table dbo.TVPItem (
  OrdNo int,
  ProdCode int,
  Qty int
)
go

-- 
-- Step 2: Create procedures to insert a single order and a single item
--

IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPOrderInsert'), 'IsProcedure') = 1
  drop procedure dbo.TVPOrderInsert
GO

create procedure dbo.TVPOrderInsert
(
	@CustCode varchar(5), 
	@OrdNo integer output, 
	@OrdDate datetime output)
as 
	set @OrdDate = GETDATE();
	insert into dbo.TVPOrd (OrdDate, CustCode) 
		values (@OrdDate, @CustCode); 
	select @OrdNo = SCOPE_IDENTITY()
go

-- old way: two procedures and multiple calls needed
IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPItemInsert'), 'IsProcedure') = 1
  drop procedure dbo.TVPItemInsert
GO

create procedure dbo.TVPItemInsert
(
	@OrdNo integer, 
	@ProdCode integer, 
	@Qty integer)
as 
	insert into dbo.TVPItem (OrdNo, ProdCode, Qty) 
		values (@OrdNo, @ProdCode, @Qty)
go

-- old way: 4 database roundtrips
declare @ordno int;
declare @orddate datetime;

exec dbo.TVPOrderInsert 'ALFKI', @ordno output, @orddate output

-- insert three lines items
exec dbo.TVPItemInsert @ordno, 111, 1
exec dbo.TVPItemInsert @ordno, 222, 2
exec dbo.TVPItemInsert @ordno, 333, 3
go
select * from dbo.TVPOrd;
select * from dbo.TVPItem;
go
		
--
-- Step 3 Create a table-valued parameter 
--        This is actually creating a user-defined type, 
--        that is a strongly typed table
--		
IF (SELECT COUNT(*) FROM sys.table_types
WHERE name = 'TVPParam' AND schema_id = 1) = 0
  create type dbo.TVPParam as table(ProdCode integer, Qty integer)
GO
		

--
-- Step 4 Redo the order entry procedure 
--        to insert 1 Order and 1-n line items in a single round trip
--
IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPOrderEntry'), 'IsProcedure') = 1
  drop procedure dbo.TVPOrderEntry
GO

create procedure dbo.TVPOrderEntry
(
	@CustCode varchar(5), 
	@Items TVPParam READONLY,  -- note that the table-valued parameter is read only
	@OrdNo integer output, 
	@OrdDate datetime output)
as 
	set @OrdDate = GETDATE();
	
	insert into TVPOrd (OrdDate, CustCode) 
		values (@OrdDate, @CustCode); 
		
	select @OrdNo = SCOPE_IDENTITY(); 
	
	insert into TVPItem (OrdNo, ProdCode, Qty) 
		select @OrdNo, ProdCode, Qty from @Items
go

-- single database roundtrip
declare @ordno int;
declare @orddate datetime;
declare @t TVPParam ;
insert into @t values(444, 4), (555,5), (666,6);  --Note use of new row contructors
exec dbo.TVPOrderEntry 'BEAUC', @t, @ordno output, @orddate output;
go
select * from dbo.TVPOrd;
select * from dbo.TVPItem; 
go

--
-- Information about the table-value parameter type is stored in system views
--
select * from sys.types				
select * from sys.table_types
go


IF OBJECTPROPERTY(OBJECT_ID('dbo.TVPFunction'), 'IsScalarFunction') = 1
  drop function dbo.TVPFunction
GO

create function TVPFunction (@t as dbo.TVPParam READONLY)
returns int
as
begin
return (select count(1) from @t)
end

go

-- Table type is member of a schema
declare @t as dbo.TVPParam

select dbo.TVPFunction(@t)  -- Note this works and returns 0 as an unassigned variable is an empty table
insert into @t values (1,1),(2,2)
select dbo.TVPFunction(@t)
go


declare @ordno int;
declare @orddate datetime;
declare @t TVPParam ;
insert into @t values(444, 4), (555,5), (666,6);
exec dbo.TVPOrderEntry 'BEAUC',@OrdNo = @ordno output, @Orddate=@orddate output; --note missing table parameter
--still works but inserts no rows
go

